{ --------------------------------------------------------------------------- }
{ DRVTYPES.PAS  Drive type detection unit.                       Version 2.12 }
{                                                                             }
{ *** PUBLIC DOMAIN ***                                                       }
{ *** NO WARRANTIES PROVIDED AND AS-IS BASIS IS ASSUMED ***                   }
{                                                                             }
{ Initially written and placed in public domain by Mr. Byte.                  }
{ Network drive detection fix, Stacker , DoubleDisk and                       }
{  Bernoully drive detection by Bobby Z.                                      }
{ RAM drive detection by Janis Smits.                                         }
{ Diskreet and SuperStor detection algorithms by Vitaly Lysenko, coded by     }
{ Mr. Byte.                                                                   }
{ --------------------------------------------------------------------------- }
{ History:

        mid November, 1994  -  initially released
         29 November, 1994  -  fixed network drive detection and added
                               Stacker drive detection
         10 December, 1994  -  added Janis's RAM drive detection method
         16 January,  1995  -  added Vertisoft DoubleDisk 2.6+ drive detection
         31 January,  1995  -  added Bernoully drive detection
         Feb-Mar,     1995  -  some idle work has been done by Mr.Byte and
                               _his_ v2.00 was released. Regrettably, he didn't
                               send new code to Bobby Z., so he continued
                               independant development.
         24 August,   1995  -  v1.21: fixed a bug with CD-ROM drive detection
         04 December, 1995  -  v1.30: changed Stacker detection, now detects
                               all versions of Stacker and is more convenient;
                               fixed Bernoulli drive detection
         29 March,    1996  -  v1.31: fixed demo program bug with removable
                               drives other than A: and B: being reported as
                               phantom floppies (I acquired a MaxOptix T3 MO
                               drive and found that it is reported as phantom
                               floppy drive 8() ).
         24 April,    1996  -  v2.10: combined version 2.0 released by Mr.Byte
                               independantly with my existing v1.31 code.
                               .10 version step is to avoid confusion with
                               v2.00 already circulating the net.
         04 December, 1997  -  v2.11: fixed incorrect detection of some RAM
                               drives as Stacker drives (additional check
                               for RAM drive made within checkStacker
                               routine).
         16 September 1999  -  v2.12: added another generic method for
                               detecting RAM disks (thanks to Tomas Navratil)
                               Now DrvTypes should properly detect more
                               RAM disks. A tradeoff is that DOS versions
                               prior to 4.00 are not supported now. I doubt
                               there still such in use though...
}

(**** General notes: ****

  This code can be easily ported to C/C++, because most of it is written in
  inline assembler. Just convert @@-style labels to __ and add "asm"s where
  needed. Note, that the call to checkStacker cannot be made in C/C++ the way
  it is done here (at least in Borland C++), you should do it like this:

  unsigned char getDriveType( unsigned char Drive )
  {
  register unsigned char temp;
  ...
  mov   temp,Drive              // passing the parameter
    }
  checkStacker(temp);
asm {
  ...
  }


 **** end of General notes ****)

{$S-,R-,I-,X+}
{ disable stack, I/O and range checking, enable extended syntax }

unit DrvTypes;
{ drive types }

interface

const
  dtError      = $00; { Invalid drive, letter not assigned }
  dtFixed      = $01; { Fixed drive }
  dtRemovable  = $02; { Removeable (floppy, etc.) drive }
  dtRemote     = $03; { Remote (network) drive }
  dtCDROM      = $04; { MSCDEX V2.00+ driven CD-ROM drive }
  dtDblSpace   = $05; { DoubleSpace compressed drive }
  dtSUBST      = $06; { SUBST'ed drive }
  dtStacker    = $07; { Stacker compressed drive }
  dtRAMDrive   = $08; { RAM drive }
  dtDublDisk   = $09; { Vertisoft DoubleDisk 2.6+ compressed drive }
  dtBernoully  = $0A; { IOmega Bernoully drive }
  dtDiskreet   = $0B; { Norton Diskreet drive }
  dtSuperStor  = $0C; { SuperStor compressed drive }

function getDriveType(Drive : Byte) : Byte;

function countValidDrives : Byte;

implementation

function checkStacker( Drive : Byte ) : Boolean; assembler;
{ returns True if Drive is Stacker compressed volume, False otherwise. }

var
        StackerDriver : Pointer;
asm
        mov     word ptr StackerDriver,0
        mov     word ptr StackerDriver[2],0
        mov     ax,4404h
        mov     bl,Drive
        mov     cx,4
        lea     dx,StackerDriver
        push    ds
        push    ss
        pop     ds
        int     21h             { get Stacker driver address }
        pop     ds
        mov     ax,word ptr StackerDriver
        or      ax,word ptr StackerDriver[2]
        jz      @@Q             { address was set? }
        push    ds           { additional check for RAM drive }
        push    ss
        pop     ds
        mov     si,sp
        sub     sp,28h       { allocate 28h bytes in stack }
        mov     dx,sp
        mov     ax,440Dh     { generic IOCTL }
        mov     cx,860h      { get device parameters }
        int     21h          { RAMDrive and VDISK don't support this command }
        mov     sp,si
        pop     ds
        jnc     @@ok
        sub     al,al
        jmp     @@Q
@@ok:
        mov     al,1
@@Q:
end; { checkStacker }

function checkDiskreet( Drive : Byte ) : Boolean; near; assembler;
{ Returns True if Drive is Norton Diskreet drive, False otherwise }
type
  TDiskreetPacket = record
    Header : array [1..6] of Byte;
    Drive  : Char;
    Size   : LongInt;
  end;
const
  DrvName  : PChar = '@DSKREET'; {-Diskreet driver name, zero-terminated string}
var
  Packet   : TDiskreetPacket;
asm
        push    ds
        mov     ax,0FE00h
        mov     di,'NU'   { 4E55h='NU' }
        mov     si,'DC'   { 4443h='DC' }
        int     2Fh
        or      al,al     { check for zero }
        je      @@2
        cmp     al,1      { check for 1 }
        je      @@2
@@1:
        sub     al,al      { return False }
        jmp     @@4
@@2:
        lds     dx,DrvName
        mov     ax,3D02h
        int     21h
        jc      @@1
        mov     bx,ax
        mov     ax,seg [Packet]
        mov     ds,ax
        mov     dx,offset [Packet]
        mov     es,ax
        mov     di,dx
        mov     cx,type TDiskreetPacket
        sub     al,al
        cld
        rep     stosb     { initialize Packet fields }
        mov     di,offset [Packet.Header]
        mov     ax,12FFh  { store first two bytes in Packet header }
        stosw
        mov     di,offset [Packet.Drive]
        mov     al,Drive
        add     al,64     { convert drive number to drive letter }
        stosb             { store drive letter }
        mov     ax,4403h  { ready to send Diskreet Packet }
        mov     cx,7
        mov     si,'dc'   { 6463h = 'dc' }
        mov     di,'NU'   { 4E55h = 'NU' }
        int     21h       { assuming ds=seg [Packet], dx=offset [Packet],
                            bx=Handle }
        mov     ah,3Eh
        int     21h       { close device }
        mov     si,offset [Packet.Size]
        lodsw
        or      ax,ax
        jnz     @@3
        lodsw
        or      ax,ax
        jz      @@1
@@3:
        mov     al,True   { return True }
@@4:
        pop     ds
end; { checkDiskreet }

function checkSuperStor( Drive : Byte ) : Boolean; near; assembler;
type
  TSSPacket = record
    Sign  : Word;
    Sign1 : Word;
    P     : Pointer;
    Res   : array [1..4] of Byte;
  end;
var
  Packet : TSSPacket;
asm
        push    ds
        mov     ax,seg [Packet]
        mov     es,ax
        mov     di,offset [Packet]
        mov     cx,type TSSPacket
        cld
        rep     stosb   { initialize SStor Packet structure }
        mov     di,offset [Packet.Sign]
        mov     ax,0AA55h
        stosw             { init Packet.Sign }
        mov     ax,0201h
        stosw             { init Packet.Sign1 }
        mov     ax,4404h
        mov     dx,seg [Packet]
        mov     ds,dx
        mov     dx,offset [Packet]
        mov     cx,12
        mov     bl,Drive
        int     21h
        jc      @@2          { if error then quit }
        mov     si,offset [Packet.Sign]
        lodsw
        or      ax,ax        { if Packet.Sign<>0 then quit }
        jnz     @@2
        lodsw
        cmp     ax,0201h     { if Packet.Sign1<>0201h then quit }
        jne     @@2
        les     di,[Packet.P]
        mov     ax,[es:di+5Dh]
        test    ax,40h       { host drive? }
        jz      @@2
        mov     cl,byte ptr es:[di+24h]
        add     cl,'A'
        mov     ah,30h
        int     21h
        cmp     ah,4
        jb      @@1
        inc     di
@@1:
        les     di,dword ptr es:[di+5Fh]
        mov     bl,[es:di]
        add     bl,'A'
        cmp     cl,Drive   { ????? I don't know whether bl or cl is a host
                             SStor drive... }
        jne     @@2
        mov     al,True   { return True }
        jmp     @@3
@@2:
        sub     al,al     { return False }
@@3:
        pop     ds
end; { checkSuperStor }

function checkRAMDrive( Drive : Byte ) : Boolean; near; assembler;
{ this function tries to access DPB for given drive and checks if drive has }
{ only 1 copy of FAT - usually this means that this is a RAM disk...        }

{ accessing DPBs without reading disk }
asm
    push    es
    push    bx
    push    cx
    mov     ah,52h      { get DOS "list of lists" pointer }
    int     21h
    les     bx,dword ptr es:[bx]  { points to first DPB }
    mov     cx,1
@@1:                    { iterate through all DPBs }
    cmp     cl,Drive
    je      @@Found
    les     bx,dword ptr es:[bx+19h] { load next DPB address }
    { ^ this is incompatible with DOS versions below 4.00, because they have }
    {   different DPB layout and next DPB pointer is at offset 18h. Peeking  }
    {   right location for this pointer depending on DOS version detected is }
    {   an option, but I'm too lazy to implement this...                     }
    inc     cx
    jmp     @@1
@@Found:
    cmp     byte ptr es:[bx+08h],1
    jne     @@Nope
    mov     al,1
    jmp     @@Q
@@Nope:
    sub     al,al
@@Q:
    pop     cx
    pop     bx
    pop     es
end;
(* alternate way - with disk access
asm
    push    ds
    push    bx
    mov     dl,Drive
    mov     ah,32h
    int     21h
    cmp     al,0FFh
    jz      @@Nope
    cmp     byte ptr ds:[bx+08h],1 { check FAT copies count }
    jnz     @@Nope
    mov     al,1
    jmp     @@Q
@@Nope:
    sub     al,al
@@Q:
    pop     bx
    pop     ds
end; *) { checkRAMDrive }


function getDriveType; assembler;
{ Detects the type for a specified drive. Drive is a drive number to detect the
  type for (0=detect current (default) drive, 1=A, 2=B, 3=C...)

  Returns: One of the dtXXX-constants.

  Note: Function will work under DOS version 3.1 or later
        Also should work under DPMI and Windows.
}

asm
        cmp     Drive,0
        jne     @@1
        mov     ah,19h    { get active drive number in al }
        int     21h
        mov     Drive,al
        inc     Drive
@@1:
        push    word ptr [Drive]
        call    checkDiskreet
        test    al,al
        jz      @@CDCheck
        mov     al,dtDiskreet
        jmp     @@Done
@@CDCheck:
        mov     ax,1500h  { check for CD-ROM v2.00+ }
        sub     bx,bx
        int     2Fh
        or      bx,bx
        jz      @@2
        mov     ax,150Bh
        sub     ch,ch
        mov     cl,Drive
        dec     cl        { !!! 1.21 - drive mapping is 0-based }
        int     2Fh
        cmp     bx,0ADADh
        jne     @@2
        or      ax,ax
        jz      @@2
        mov     bl,dtCDROM
        jmp     @@Done
@@2:
        mov     ax,4409h { check for SUBST'ed drive }
        mov     bl,Drive
        int     21h
        jc      @@s
        test    dh,80h
        jz      @@s
        mov     bl,dtSUBST
        jmp     @@Done
@@s:
        mov     ax,4A11h  { check for DoubleSpace drive }
        mov     bx,1
        mov     dl,Drive
        dec     dl
        int     2Fh
        or      ax,ax     { is DoubleSpace loaded? }
        jnz     @@network
        cmp     dl,bl     { if a host drive equal to compressed, then get out... }
        je      @@network
        test    bl,80h    { bit 7=1: DL=compressed,BL=host
                                 =0: DL=host,BL=compressed }
        jz      @@SStorChk  { so avoid host drives, assume host=fixed :) }
        inc     dl
        cmp     Drive,dl
        jne     @@SStorChk
        mov     bl,dtDblSpace
        jmp     @@Done
@@SStorChk:
        push    word ptr [Drive]
        call    checkSuperStor
        test    al,al
        jz      @@network
        mov     al,dtSuperStor
        jmp     @@Done
@@network:
        mov     ax,4409h     { check for remote drive }
        mov     bl,Drive
        int     21h
        jc      @@5
        and     dh,10h
        jz      @@4
        mov     bl,dtRemote
        jmp     @@Done
@@4:
        push    word ptr [Drive] { check for Stacker volume }
        call    checkStacker
        test    al,al
        jz      @@8
        mov     bl,dtStacker
        jmp     @@Done
@@8:
        push    word ptr [Drive]
        call    checkRAMDrive
        test    al,al
        jz      @@checkHD
        mov     bl,dtRAMDrive
        jmp     @@Done
@@checkHD:
        { we'll do an additional check for RAMdrive later on }
        mov     ax,4408h     { check for fixed (hard) drive }
        mov     bl,Drive
        int     21h
        jc      @@5
        test    al,al        { 0 means removable media }
        jz      @@check_Bernoulli
        push    ds           { check for RAM drive }
        push    ss
        pop     ds
        mov     si,sp
        sub     sp,28h       { allocate 28h bytes in stack }
        mov     dx,sp
        mov     ax,440Dh     { generic IOCTL }
        mov     cx,860h      { get device parameters }
        int     21h          { RAMDrive and VDISK don't support this command }
        jc      @@cleanup
        pushf
        mov     di,dx
        cmp     byte ptr ds:[di+6],0F8h { DoubleDisk returns 0F8h in media type}
        jz      @@dubldsk               { field of BPB if drive in question is }
                                        { compressed }
        popf
        jmp     @@cleanup
@@dubldsk:
        popf
        mov     bl,dtDublDisk
        mov     sp,si
        pop     ds
        jmp     @@Done
@@cleanup:
        mov     sp,si
        pop     ds
@@RAMDrive:
        mov     bl,dtRAMDrive
        jc      @@Done
        jmp     @@fixed                 { if we get here - it's fixed drive }
@@check_Bernoulli:
        cmp     Drive,2
        jbe     @@6                     { drives A: and B: are floppies, avoid }
        push    ds                      { checking them }
        mov     ah,1Ch                  { this function works _really_ slowly }
        mov     dl,Drive                { get media descriptor pointer }
        int     21h
        cmp     byte ptr ds:[bx],0FDh   { 360K diskette? }
        pop     ds
        jnz     @@6                     { just a removable drive }
        push    ds
        mov     ah,32h                  { get BPB pointer }
        mov     dl,Drive
        int     21h
        cmp     byte ptr ds:[bx+0Bh],2  { Sectors per FAT is more than 2 for }
        pop     ds                      { Bernoully drives }
        jz      @@6
        mov     bl,dtBernoully
        jmp     @@Done
@@fixed:
        mov     bl,dtFixed
        jmp     @@Done
@@5:
        sub     bl,bl        { mov bl,dtError cuz dtError=0 }
        jmp     @@Done
@@6:
        mov     bl,dtRemovable   { else - removeable media }
@@Done:
        mov     al,bl
end; { getDriveType }

function countValidDrives;
{ - returns number of assigned letters in system }
 var i : Byte;
     C : Byte;
 begin
  C := 0;
  for i := 1 to 26 do
   if GetDriveType(i) <> dtError then
    Inc(C);
  countValidDrives := C;
 end; { countValidDrives }

end.
